(*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON-Web.
 *
 * UMLMON-Web is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON-Web is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON-Web; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: common.ml 268 2010-08-02 00:33:37Z gerd $
 * ----------------------------------------------------------------------
 * Common RPC-related functions
 *)
 
open Wd_dialog
open Wd_run_cgi
open Wd_types
open Netcgi
open Umlmonctrl_aux

module Mon = Umlmonctrl_clnt.Ctrl.V1
module Dir = Umlmonctrl_clnt.Directory.V1

exception Wrong_password
exception Remote_error of string
exception Interrupted


let get_creds dtd password =
  let pw1 =
    try
      let mpw_ent, _ =
	dtd # gen_entity "master-password" in
      let mpw_str =
	Pxp_dtd.Entity.replacement_text mpw_ent in
      if mpw_str = "" then
	[]
      else
	[mpw_str]
    with Pxp_types.WF_error _ -> [] in
  let pw2 =
    if password = "" then [] else [ password ] in
  pw1 @ pw2



let rec authenticate ?userpw client creds =
  (* Note:  Umlmon_client.authenticate groks up to two passwords at a time *)
  match creds with
    | pw1 :: pw2 :: creds' ->
	let ok = Umlmon_client.authenticate ?userpw client [pw1;pw2] in
	if not ok then authenticate ?userpw client creds'
    | [pw] ->
	let ok = Umlmon_client.authenticate ?userpw client [pw] in
	if not ok then raise Wrong_password
    | [] ->
	raise Wrong_password
;;


let rec authenticate'async ?userpw client creds action =
  (* Note:  Umlmon_client.authenticate groks up to two passwords at a time *)
  match creds with
    | pw1 :: pw2 :: creds' ->
	Umlmon_client.authenticate'async ?userpw client [pw1;pw2]
	  (fun get_reply ->
	     let ok = try get_reply() with _ -> false in
	     if ok then
	       action (fun () -> ())
	     else
	       authenticate'async ?userpw client creds' action
	  )
    | [pw] ->
	Umlmon_client.authenticate'async ?userpw client [pw]
	  (fun get_reply ->
	     let ok = try get_reply() with _ -> false in
	     if ok then
	       action (fun () -> ())
	     else
	       action (fun () -> raise Wrong_password)
	  )
    | [] ->
	action (fun () -> raise Wrong_password)
;;


let check_auth ?userpw id creds =
  let ues = Unixqueue.create_unix_event_system() in
  let client = Umlmon_client.create_client 
                  ~prefer_local_socket:true ~prefer_redirection:true ues id in
   try
     authenticate ?userpw client creds;
     Rpc_client.shut_down client;
   with
       any ->        
	 Rpc_client.shut_down client;
	 raise any
;;


let get_status id creds =
   let ues = Unixqueue.create_unix_event_system() in
   try
     let client = Umlmon_client.create_client 
                      ~prefer_local_socket:true ~prefer_redirection:true ues id in
     try
       authenticate client creds;
       let code = 
	 ( try
	     match Mon.get_status client () with
	       | `ctrl_inactive -> "INACTIVE"
	       | `ctrl_stopped  -> "STOPPED"
	       | `ctrl_running  -> "RUNNING"
	       | `ctrl_starting -> "STARTING"
	   with
	       _ -> "NA") in
       Rpc_client.shut_down client;
       code
     with
	 err -> Rpc_client.shut_down client; raise err
   with
     | Wrong_password as exn -> 
	 Log.log_error ("Wrong password for VM: " ^ id);
	 "NA"
     | _ -> "NA"
;;


let get_status_in_parallel id_list creds =
  let ues = Unixqueue.create_unix_event_system() in
  let client_list =
    List.map
      (fun id ->
	 try
	   Some(id, Umlmon_client.create_client 
		     ~prefer_local_socket:true  ~prefer_redirection:true ues id)
	 with
	     _ -> None)
      id_list in
  let results = Hashtbl.create 100 in
  List.iter
    (function
       | Some (id,client) ->
	   authenticate'async client creds
	     (fun get_result ->
		try
		  let () = get_result() in   (* or Wrong_password or ... *)
		  Mon.get_status'async client ()
		    (fun get_result ->
		       try
			 let s = get_result() in   (* or exception *)
			 let code =
			   match s with
			     | `ctrl_inactive -> "INACTIVE"
			     | `ctrl_stopped  -> "STOPPED"
			     | `ctrl_running  -> "RUNNING"
			     | `ctrl_starting -> "STARTING" in
			 Hashtbl.add results id code
		       with
			   _ -> Hashtbl.add results id "NA"
		    )
		with
		    _ -> Hashtbl.add results id "NA"
	     )
       | None ->
	   ()
    )
    client_list;
  (* Now do everything: *)
  Unixqueue.run ues;
  (* Clean up: *)
  List.iter
    (function
       | Some (id,client) -> Rpc_client.shut_down client
       | None -> ()
    )
    client_list;
  (* Collect results: *)
  List.map
    (fun id -> try Hashtbl.find results id with Not_found -> "NA")
    id_list
;;


let get_monitors_in_parallel host_list =
  let ues = Unixqueue.create_unix_event_system() in
  let client_list =
    List.map
      (fun host ->
	 try
	   Some(host, Umldir_client.create_client ues host)
	 with
	     _ -> None)
      host_list in
  let results = ref [] in
  List.iter
    (function
       | Some (host,client) ->
	   Dir.get_monitors'async client ()
	     (fun get_result ->
		try
		  let l = Array.to_list (get_result()) in  (* or exception *)
		  let id_list =
		    List.map (fun mon -> mon.monitor_id ^ "@" ^ host) l in
		  results := id_list @ !results;
		with
		    _ -> ()
	     )
       | None -> ()
    )
    client_list;
  (* Now do everything: *)
  Unixqueue.run ues;
  (* Clean up: *)
  List.iter
    (function
       | Some (_,client) -> Rpc_client.shut_down client
       | None -> ()
    )
    client_list;
  (* Return results: *)
  !results
;;


let call_proc id creds f arg =
  let ues = Unixqueue.create_unix_event_system() in
  let client = Umlmon_client.create_client 
                  ~prefer_local_socket:true ~prefer_redirection:true ues id in
  try
    authenticate client creds;
     let r = f client arg in
     Rpc_client.shut_down client;
     r
  with
    | err ->
	Rpc_client.shut_down client;
	raise err
;;


let call_simple_proc id creds f arg =
  match call_proc id creds f arg with
    | `code_ok ->
	()
    | `code_error msg ->
	raise (Remote_error msg)
    | `code_interrupt ->
	raise Interrupted
    | `code_in_progress -> 
	()
