(*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON-Web.
 *
 * UMLMON-Web is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON-Web is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON-Web; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: termout.ml 268 2010-08-02 00:33:37Z gerd $
 * ----------------------------------------------------------------------
 *
 *)

open Umlmonctrl_aux
open Printf

module Mon = Umlmonctrl_clnt.Ctrl.V1

let rec microsleep t =
  let t0 = Unix.gettimeofday() in
  try
    ignore(Unix.select [] [] [] t)
  with
      Unix.Unix_error(Unix.EINTR,_,_) ->
        microsleep (t -. (Unix.gettimeofday() -. t0))
;;


let config = lazy (
  let config_xml = 
    if Sys.file_exists "./umlmon_web-config.xml" then
      "./umlmon_web-config.xml" 
    else
      Filename.concat Instconf.prefix "share/umlmon_web/ui/umlmon_web-config.xml"
  in
  let pxp_config = Pxp_yacc.default_config in
  Pxp_yacc.parse_dtd_entity pxp_config (Pxp_yacc.from_file config_xml)
)
;;


let output_header (cgi : Netcgi.cgi_activation) html =
  let sup_ent, _ =
    (Lazy.force config) # gen_entity "static-url-prefix" in
  let sup_str =
    Pxp_dtd.Entity.replacement_text sup_ent in
  cgi # output # output_string "<html>\n";
  cgi # output # output_string "<head>\n";
  cgi # output # output_string ("<link type=\"text/css\" rel=\"stylesheet\" href=\"" ^ sup_str ^ "definitions.css\">\n");
  cgi # output # output_string "</head>\n";
  cgi # output # output_string "<body class='termout'>\n";
  cgi # output # output_string "<pre class=\"console\" id=\"console\">\n";
  cgi # output # output_string html;
  cgi # output # output_string "</pre>\n";
  cgi # output # commit_work()
;;


let generate_html s =
  let i4 =  Rtypes.int_of_int4 in
  let u4 =  Rtypes.int_of_uint4 in
  let color atts mask shift =
    match (atts land (u4 mask)) lsr (u4 shift) with
      | n when n = i4 c_black -> "black"
      | n when n = i4 c_red -> "red"
      | n when n = i4 c_green -> "green"
      | n when n = i4 c_brown -> "brown"
      | n when n = i4 c_blue -> "blue"
      | n when n = i4 c_magenta -> "magenta"
      | n when n = i4 c_cyan -> "cyan"
      | n when n = i4 c_white -> "white"
      | _ -> ""
  in

  let b = Buffer.create 1024 in
  for y_pos = 0 to (Int32.to_int s.te_height) - 1 do
    let last_atts = ref None in
    for x_pos = 0 to (Int32.to_int s.te_width) - 1 do
      let cell = s.te_screen.(y_pos).(x_pos) in
      let data = cell.cell_char in
      let atts = Int32.to_int cell.cell_atts in
      if !last_atts <> Some atts then (
	( match !last_atts with
	    | None -> ()
	    | Some _ -> 
		Buffer.add_string b "</span>"
	);
	let reverse =
	  match (atts land (u4 mask_reverse)) lsr (u4 shift_reverse) with
	    | n when n = i4 r_on -> true
	    | _ -> false in
	let classes =
	  ( match (atts land (u4 mask_intensity)) lsr (u4 shift_intensity) with
	      | n when n = i4 i_norm -> []
	      | n when n = i4 i_dim  -> [ "t-dim" ]
	      | n when n = i4 i_bold -> [ "t-bold" ]
	      | _ -> [] ) @
	  ( match (atts land (u4 mask_underscore)) lsr (u4 shift_underscore) with
	      | n when n = i4 u_on -> [ "t-underscore" ]
	      | _ -> [] ) @
	  ( match (atts land (u4 mask_blink)) lsr (u4 shift_blink) with
	      | n when n = i4 b_on -> [ "t-blink" ]
	      | _ -> [] ) @
	  ( if reverse then
	      [ "t-fg-" ^ color atts mask_bg_color shift_bg_color; 
	        "t-bg-" ^ color atts mask_fg_color shift_fg_color ]
	    else
	      [ "t-fg-" ^ color atts mask_fg_color shift_fg_color; 
	        "t-bg-" ^ color atts mask_bg_color shift_bg_color ] )
	in
	Buffer.add_string b (sprintf "<span class='%s'>" (String.concat " " classes));
      );
      last_atts := Some atts;
	let is_cursor_pos = 
	  s.te_cursor_visible &&
	  x_pos = Int32.to_int s.te_cursor_x && 
	  y_pos = Int32.to_int s.te_cursor_y in
	if is_cursor_pos then
	  Buffer.add_string b "<span id='t-cursor'>";
	( match data with
	      (-1l) -> Buffer.add_string b "&nbsp;"
	    | 32l   -> Buffer.add_string b "&nbsp;"
	    | n -> 
		if (n > 32l && n <= 127l) || n >= 160l then
		  Buffer.add_string b ("&#" ^ Int32.to_string n ^ ";")
	);
	if is_cursor_pos then
	  Buffer.add_string b "</span>";
    done;
    Buffer.add_string b "</span><br>"
  done;
  Buffer.contents b
;;


let split_up s =
  let l = String.length s in
  let rec loop k =
    if k < l then
      let n = min 80 (l - k) in
      String.sub s k n :: loop (k+n)
    else
      []
  in
  loop 0
;;


let output_update (cgi : Netcgi.cgi_activation) s =
  let html = generate_html s in
  let html_l = split_up html in
  let html' = String.concat "\";\n html += \"" html_l in
  cgi # output # output_string "<script type='text/javascript'>\n";
  cgi # output # output_string "var html = \"";
  cgi # output # output_string html';
  cgi # output # output_string "\";\n";
  cgi # output # output_string "document.getElementById('console').innerHTML = html;\n";
  cgi # output # output_string "</script>\n";
  cgi # output # output_string "<!-- Make IE happy -->\n";
  cgi # output # output_string "<!-- Make IE happy -->\n";
  cgi # output # output_string "<!-- Make IE happy -->\n";
  cgi # output # output_string "<!-- Make IE happy -->\n";
  cgi # output # output_string "<!-- Make IE happy -->\n";
  cgi # output # output_string "<!-- Make IE happy -->\n";
  cgi # output # output_string "<!-- Make IE happy -->\n";
  cgi # output # output_string "<!-- Make IE happy -->\n";
  cgi # output # output_string "<!-- Make IE happy -->\n";
  cgi # output # output_string "<!-- Make IE happy -->\n";
  cgi # output # output_string "<!-- Make IE happy -->\n";
  cgi # output # output_string "<!-- Make IE happy -->\n";
  cgi # output # output_string "<!-- Make IE happy -->\n";
  cgi # output # output_string "<!-- Make IE happy -->\n";
  cgi # output # output_string "<!-- Make IE happy -->\n";
  cgi # output # output_string "<!-- Make IE happy -->\n";
  cgi # output # commit_work()
;;


let call_simple_proc client f arg =
  match f client arg with
    | `code_ok ->
        ()
    | `code_error msg ->
        raise (Common.Remote_error msg)
    | `code_interrupt ->
        raise Common.Interrupted
    | `code_in_progress ->
        ()

let process (cgi : Netcgi.cgi_activation) =
  (* Get parameters: *)
  let id = cgi # argument_value "id" in
  let password = cgi # argument_value "password" in
  let ct_name = cgi # argument_value "ct_name" in
  let ct_id = cgi # argument_value "ct_id" in
  let handle = { ct_name = ct_name; ct_id = Int32.of_string ct_id } in

  cgi # set_header();

  try
    let ues = Unixqueue.create_unix_event_system() in
    let client = Umlmon_client.create_client 
                     ~prefer_local_socket:true ~prefer_redirection:true ues id in

    let creds = Common.get_creds (Lazy.force config) password in
    Common.authenticate client creds;

    (* Init output: *)
    let t = ref (Unix.gettimeofday()) in
    let html =
      match Mon.te_state client handle with
	| None -> 
	    failwith "Diese Konsole hat keinen Zugriff (mehr) auf den Terminal-Emulator."
	| Some s ->
	    generate_html s
    in
    output_header cgi html;
    
    (* Wait for arriving data *)
    while true do
      try
	call_simple_proc client Mon.te_read (handle, 10l);
	let t_now = Unix.gettimeofday() in
	microsleep (max 0.0 (0.2 -. (t_now -. !t)));
	t := Unix.gettimeofday();
	match Mon.te_state client handle with
	  | None -> 
	      failwith "Diese Konsole hat keinen Zugriff (mehr) auf den Terminal-Emulator."
	  | Some s ->
	      output_update cgi s;
      with
	| Common.Interrupted ->
	    cgi # output # output_string "<!-- Keep alive -->\n";
	    cgi # output # commit_work()
    done
  with
    | Common.Wrong_password ->
	cgi # output # output_string ("Falsches Passwort!");
	cgi # output # commit_work()
    | Common.Remote_error msg ->
	cgi # output # output_string ("UMLMON-Fehler: " ^ msg);
	cgi # output # commit_work()
    | Failure msg ->
	cgi # output # output_string ("Lokaler Fehler: " ^ msg);
	cgi # output # commit_work()
;;

