(*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON.
 *
 * UMLMON is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: umlwatch.ml 253 2007-06-29 14:05:31Z gerd $ *)

open Umlmonctrl_aux
open Umlmon_client
open Printf

module Cl = Umlmonctrl_clnt.Ctrl.V1


exception Cannot_connect_term of string


let ues = Unixqueue.create_unix_event_system();;

let del_ctrl_re = Str.regexp "[\000-\031\127]";;

let tty_setup ?(interrupt=false) ?(crnl=false) f =
  let attr = Unix.tcgetattr Unix.stdin in
  let attr' = 
    { attr with
	Unix.c_ignbrk = not interrupt; (* maybe no CTRL-C = interrupt *)
	c_inlcr = false;      (* No LF -> CR mapping on input *)
	c_icrnl = crnl;       (* Maybe no CR -> LF mapping on input *)
	c_igncr = false;      (* Keep CR *)
	c_istrip = false;     (* 8 bit input *)
	c_ixon = false;       (* No XON/XOFF for input *)
	c_ixoff = false;      (* No XON/XOFF for input *)

	c_opost = false;      (* No output processing *)

	c_isig = interrupt;   (* No keyboard signals *)
	c_icanon = false;     (* No canonical mode *)
	c_echo = false;       (* Disable echo *)
	c_echoe = false;
	c_echok = false;
	c_echonl = false;

	c_vmin = 1;
	c_vtime = 0;
    } in
  Unix.tcsetattr Unix.stdin Unix.TCSANOW attr';
  try
    f();
    Unix.tcsetattr Unix.stdin Unix.TCSANOW attr
  with
      err ->
	Unix.tcsetattr Unix.stdin Unix.TCSANOW attr;
	raise err
;;


let contact id =
  let client = create_client ~prefer_redirection:true ues id in
  if need_authentication client then (
    prerr_string "Password: ";
    flush stderr;
    tty_setup 
      ~interrupt:true
      ~crnl:true
      (fun () ->
	 let password = read_line() in
	 prerr_endline "\r";
	 flush stderr;
	 if not (authenticate client [password]) then
	   raise (Rpc.Rpc_server Rpc.Auth_failed);
      );
  );
  client
;;


let ubd_cow_re = Str.regexp "^\\([^,]+\\),\\(.+\\)$"


let info cl =
  let vm_id = Cl.get_id cl () in
  let vm_status = Cl.get_status cl () in
  let vm_terms = Cl.get_monitored_terminals cl () in
  let config = Cl.config_get cl () in
  let disks = Cl.disk_files cl () in
  let kernels = Cl.announced_kernels cl () in

  let star_line = String.make 79 '*' in
  let minus_line = "+" ^ String.make 77 '-' ^ "+" in

  printf "%s\n" star_line;
  printf "* %-75s *\n" ("Status for VM instance " ^ vm_id);
  printf "%s\n\n" star_line;

  printf "%s\n" minus_line;
  printf "| %-75s |\n" "Configuration";
  printf "%s\n" minus_line;
  List.iter
    (fun param ->
       printf "| %13s = %-59s |\n" param.param_name param.param_value
    )
    ( Array.to_list config.cfg_privsetup @
      [ config.cfg_start ] @
      ( match config.cfg_logfile with
	  | None -> []
	  | Some p -> [p]
      ) @
      [ config.cfg_kernel ] @
      Array.to_list config.cfg_kernelargs @
      [ config.cfg_mem ] @
      Array.to_list config.cfg_con @
      Array.to_list config.cfg_ssl @
      Array.to_list config.cfg_ubd @
      Array.to_list config.cfg_eth @
      Array.to_list config.cfg_extension
    );
  printf "%s\n\n" minus_line;

  if kernels <> [| |] then (
    printf "%s\n" minus_line;
    printf "| %-75s |\n" "Kernels";
    printf "%s\n" minus_line;
    printf "| %-75s |\n" "KERNEL                        FILE";
    Array.iter
      (fun k ->
	 printf "| %-29s %45s |\n" k.kernel_name k.kernel_file
      )
      kernels
  );

  if disks <> [| |] then (
    printf "%s\n" minus_line;
    printf "| %-75s |\n" "Disks";
    printf "%s\n" minus_line;
    printf "| %-75s |\n" "DEVICE TYP   STATUS       SIZE FILE(S)";

    let ubd_list = Array.to_list config.cfg_ubd in

    let ext_disks = 
      Array.map
	(fun disk ->
	   try
	     let ubd =
	       List.find
		 (fun ubd ->
		    if Str.string_match ubd_cow_re ubd.param_value 0 then
		      let cow_file = Str.matched_group 1 ubd.param_value in
		      cow_file = disk.disk_name 
		    else
		      ubd.param_value = disk.disk_name
		 )
		 ubd_list in
	     (Some ubd.param_name, disk)
	   with
	       Not_found -> 
		 (None, disk)
	)
	disks in
    
    Array.sort
      (fun x y ->
	 match (x,y) with
	   | ((Some x_device, x_disk), (Some y_device, y_disk)) ->
	       if x_device = y_device then
		 compare x_disk.disk_name y_disk.disk_name
	       else
		 compare x_device y_device
	   | ((None, x_disk), (None, y_disk)) ->
	       compare x_disk.disk_name y_disk.disk_name
	   | ((None, _), (Some _, _)) ->
	       1
	   | ((Some _, _), (None, _)) ->
	       (-1)
      )
      ext_disks;
    
    Array.iter
      (fun (device_opt, disk) ->
	 printf "| %6s %3s %9s %9s %-44s |\n"
	   (match device_opt with Some d -> d | None -> "")
	   (if disk.disk_is_cow then "COW" else "DSK")
	   ( match disk.disk_info with
	       | `disk_exists info -> 
		   if info.disk_use_count > 0l then 
		     "BACKING" 
		   else 
		     if info.disk_lock = `lock_write then
		       "LOCKED"  (* Our own write locks are not shown *)
		     else
		       let access = 
			 if not info.disk_rw then
			   "ro"
			 else
			   match info.disk_lock with
			     | `lock_read
			     | `lock_read_own -> "sh"  (* = shared *)
			     | _ -> "wr" 
		       in
		       "AVAIL(" ^ access ^ ")"
	       | `disk_being_created -> "CREATING"
	       | `disk_error msg -> "ERROR"
	   )
	   ( match disk.disk_info with
	       | `disk_exists info -> 
		   Int64.to_string (Int64.div info.disk_size 1048576L) ^ "M"
	       | _ -> ""
	   )
	   disk.disk_name;
       
	 ( match disk.disk_info with
	     | `disk_exists info ->
		 ( match info.disk_backing_file with
		     | None -> ()
		     | Some bf -> 
			 printf "| %30s %-44s |\n" "" ("backing: " ^ bf)
		 );
	     | `disk_being_created -> ()
	     | `disk_error msg ->
		 printf "|     %-71s |\n" msg
	 );
      )
      ext_disks;

    printf "%s\n\n" minus_line;
  );

  if vm_terms <> [| |] then (
    printf "%s\n" minus_line;
    printf "| %-75s |\n" "Terminals";
    printf "%s\n" minus_line;
    printf "| %-75s |\n" "DEVICE TYP CONN LOGFILE";

    Array.sort (fun x y -> compare x.ctrl_termname y.ctrl_termname) vm_terms;

    Array.iter
      (fun vm_term ->
	 match vm_term.ctrl_termdef with
	   | `ctrl_pty pty ->
	       printf "| %6s %3s %4s %-59s |\n"
		 vm_term.ctrl_termname
		 "PTY"
		 ( if pty.ctrl_connected then "yes" else "" )
		 ( match pty.ctrl_logfile with
		     | None -> ""
		     | Some name -> name
		 )
      )
      vm_terms;
    printf "%s\n\n" minus_line;
  );
  
  printf "%s\n" minus_line;
  printf "| %-75s |\n" 
    ( "VM status: " ^ 
      match vm_status with
	| `ctrl_inactive -> "inactive"
	| `ctrl_stopped -> "stopped"
	| `ctrl_running -> "running"
	| `ctrl_starting -> "starting" );
  printf "%s\n" minus_line;
  flush stdout;
;;


let do_proc f arg cl =
  match f cl arg with
    | `code_ok -> ()
    | `code_in_progress -> ()
    | `code_error msg ->
	failwith ("Server responds with error: " ^ msg)
    | `code_interrupt ->
	failwith "Interrupt";
;;


let do_connwith arg cl =
  match Cl.connect_with cl arg with
    | `code_ok _ -> ()
    | `code_in_progress -> ()
    | `code_error msg ->
	failwith ("Server responds with error: " ^ msg)
    | `code_interrupt ->
	failwith "Interrupt";
;;


let check_get_set_list config =
  [ Str.regexp "^con[0-9]+$", 
    (fun () -> config.cfg_con),
    (fun x -> { config with cfg_con = x });
    
    Str.regexp "^ssl[0-9]+$",
	      (fun () -> config.cfg_ssl),
	      (fun x -> { config with cfg_ssl = x });
    
    Str.regexp "^ubd[0-9]+$",
    (fun () -> config.cfg_ubd),
    (fun x -> { config with cfg_ubd = x });
    
    Str.regexp "^eth[0-9]+$",
    (fun () -> config.cfg_eth),
    (fun x -> { config with cfg_eth = x }) ]
  @
    (List.map
       (fun cfg -> (Str.regexp (Str.quote cfg.param_name),
		    (fun () -> config.cfg_extension),
		    (fun x -> { config with cfg_extension = x })))
       (Array.to_list config.cfg_extension))
;;


let set_parameter (name,value) cl =
  let param n v = { param_name = n; param_value = v } in
  let config = Cl.config_get cl () in
  let config' =
    match name with
      | "logfile" ->
	  { config with cfg_logfile = Some (param "logfile" value) }
      | "kernel" ->
	  { config with cfg_kernel = param "kernel" value }
      | "kernelarg" ->
	  { config with 
	      cfg_kernelargs = 
	        Array.append config.cfg_kernelargs [| param "kernelarg" value |] }
      | "mem" ->
	  { config with cfg_mem = param "mem" value }
      | "start" ->
	  { config with cfg_start = param "start" value }
      | _ ->
	  ( try
	      let _, get, set =
		List.find (fun (check, get, set) ->
			     Str.string_match check name 0)
		  (check_get_set_list config) in
	      
	      let old_params = Array.to_list (get ()) in
	      let new_params =
		List.filter (fun p -> p.param_name <> name) old_params @
		[ param name value ] in
	      set (Array.of_list new_params)

	    with
		Not_found ->
		  failwith ("Unknown parameter: " ^ name)
	  )
  in
  do_proc Cl.config_set { config' with cfg_privsetup = [| |] } cl
;;


let rec but_last l = 
  match l with
    | [] -> raise Not_found
    | [x] -> []
    | x :: l' -> x :: but_last l'


let del_parameter name cl =
  let param n v = { param_name = n; param_value = v } in
  let config = Cl.config_get cl () in
  let config' =
    match name with
      | "logfile" ->
	  { config with cfg_logfile = None }
      | "kernel" ->
	  failwith "Cannot delete this parameter"
      | "kernelarg" ->
	  { config with 
	      cfg_kernelargs = ( try Array.of_list 
				   (but_last (Array.to_list config.cfg_kernelargs))
				with Not_found -> [| |] ) };
      | "mem" ->
	  failwith "Cannot delete this parameter"
      | "start" ->
	  failwith "Cannot delete this parameter"
      | _ ->
	  ( try
	      let _, get, set =
		List.find (fun (check, get, set) ->
			     Str.string_match check name 0)
		  (check_get_set_list config) in
	      
	      let old_params = Array.to_list (get ()) in
	      let new_params =
		List.filter (fun p -> p.param_name <> name) old_params in
	      set (Array.of_list new_params)

	    with
		Not_found ->
		  failwith ("Unknown parameter: " ^ name)
	  )
  in
  do_proc Cl.config_set { config' with cfg_privsetup = [| |] } cl
;;


let connect cl termid =
  (* Connect terminal *)
  let handle =
    match Cl.connect cl termid with
      | `code_ok h ->  h
      | _ -> raise (Cannot_connect_term termid)
  in

  print_string "You can disconnect by typing ^]\n";
  flush stdout;

  (* Setup our own tty: *)
  tty_setup
    (fun () ->
       let connected = ref true in
       
       (* Arrange that all data from stdin are sent to the monitor: *)
       let rd_buf = String.create 8192 in

       let cur_in_eng = ref None in

       let rec receive_from_stdin() =
	 let in_eng =
	   new Uq_engines.poll_engine
	     [ Unixqueue.Wait_in Unix.stdin, (-1.0) ]
	     ues in
	 cur_in_eng := Some in_eng;
	 
	 Uq_engines.when_state 
	   ~is_done:(fun _ ->
		       let n = Unix.read Unix.stdin rd_buf 0 (String.length rd_buf) in
		       if n = 0 then raise End_of_file;  (* How can this happen ? *)
		       let s = String.sub rd_buf 0 n in
		       Cl.write'async cl (handle,s,10l) (fun _ -> ());
		       if s = "\029" then (
			 (* Disconnect: *)
			 connected := false;
			 cur_in_eng := None;
			 Cl.disconnect'async cl (handle,false) (fun _ -> ());
		       )
		       else
			 receive_from_stdin()
		  )
	   in_eng;
       in
       receive_from_stdin();

       (* Arrange that all data coming from the monitor are printed to stdout: *)

       let rec receive_from_mon () =
	 Cl.read'async cl (handle,10l)
	   (fun get_result ->
	      ( match get_result() with
		  | `code_ok data ->
		      print_string data;
		      flush stdout;
		  | `code_error m ->
		      (* Currently this can only happen if we are disconnected by
		       * the monitor:
		       *)
		      ( match !cur_in_eng with
			  | None -> ()
			  | Some eng -> eng # abort(); cur_in_eng := None
		      );
		      connected := false
		  | `code_in_progress ->
		      ()
		  | `code_interrupt ->
		      ()
	      );
	      if !connected then receive_from_mon()
	   )
       in
       receive_from_mon();

       (* Now do everything until you are happy: *)

       Unixqueue.run ues;
    );
  prerr_endline "\nDisconnect.\n";
  Cl.disconnect'async cl (handle,false) (fun _ -> ());  (* Just to be sure. *)
;;


let force_connect cl termid =
  ignore(Cl.disconnect cl ({ ct_name = termid; ct_id = 0l }, true) );
  connect cl termid
;;


let vm_wait n cl =
  let counter = ref 0 in
  while Cl.get_status cl () <> `ctrl_inactive do
    Unix.sleep 1;
    incr counter;
    if !counter >= n then failwith "VM has not yet finished"
  done
;;


let show_port cl =
  match Cl.port cl () with
    | None ->
	print_endline "none"
    | Some p ->
	print_endline (Int32.to_string p)
;;


let change_pw id =
  let old_pw = ref "" in
  let new_pw = ref "" in
  let new_pw_again = ref "" in
  prerr_string "Old user password: ";
  flush stderr;
  tty_setup 
    ~interrupt:true
    ~crnl:true
    (fun () ->
       old_pw := read_line()
    );
  prerr_endline "";
  (* We always create a new client, such that the old client is still active on
   * auth errors.
   *)
  let cl = create_client ~prefer_redirection:true ues id in
  if not (authenticate ~userpw:true cl [!old_pw]) then
    failwith "Bad password";
  prerr_string "New user password: ";
  flush stderr;
  tty_setup 
    ~interrupt:true
    ~crnl:true
    (fun () ->
       new_pw := read_line()
    );
  prerr_endline "";
  prerr_string "Retype new user password: ";
  flush stderr;
  tty_setup 
    ~interrupt:true
    ~crnl:true
    (fun () ->
       new_pw_again := read_line()
    );
  prerr_endline "";
  if !new_pw <> !new_pw_again then
    failwith "This is not the same password!";
  change_password cl !old_pw !new_pw


let tail (file,n) cl =
  let r = Cl.logfile_read cl (file, Int32.neg n, n) in
  match r with
    | `code_ok stat ->
	Array.iter
	  (fun s -> print_endline (Str.global_replace del_ctrl_re "" s))
	  stat.lines;
    | `code_error msg ->
	failwith ("Server responds with error: " ^ msg)
    | `code_in_progress ->
	failwith "No result"
    | `code_interrupt ->
	failwith "Interrupt"
;;


let site_cmd args cl =
  let r = Cl.site_command cl (Array.of_list (Str.split (Str.regexp "[ \t\r\n]+") args)) in
  match r with
    | `code_ok msg ->
	printf "Command executed successfully\n";
	print_endline msg;
	flush stdout
    | `code_error msg ->
	failwith ("Error: " ^ msg)
    | `code_in_progress ->
	failwith "No result"
    | `code_interrupt ->
	failwith "Interrupt"
;;


let disk_create_re = Str.regexp "\\([^:]+\\):\\([0-9]+\\)\\(:sparse\\)?$"

let disk_resize_re = Str.regexp "\\([^:]+\\):\\([0-9]+\\)\\(:sparse\\)?$"

let disk_create_cow_re = Str.regexp "\\([^:]+\\):\\([^:]+\\)$"

let disk_copy_re = Str.regexp "\\([^:]+\\):\\([^:]+\\)$"

let set_re = Str.regexp "\\([a-zA-Z0-9_]+\\)=\\(.*\\)$"

let connwith_re = Str.regexp "\\([a-zA-Z0-9_]+\\)=\\(.*\\):\\(r?w?\\)$"

let tail_re = Str.regexp "\\([a-zA-Z0-9_]+\\)\\(,\\([0-9]+\\)\\)?$"


let version() =
  printf "UMLMON Release %s Revision %s\n" Instconf.release Instconf.revision;
  printf "Copyright 2005 by Informatikbuero Gerd Stolpmann.\n\n";
  printf "UMLMON is free software, and can be distributed under the terms\n";
  printf "of the GNU General Public License.\n\n";
  printf "Commercial support available, please contact gerd@gerd-stolpmann.de\n";
  flush stdout
;;


let command_help spec =
  let b = Buffer.create 1000 in
  Buffer.add_string b "List of commands:\n";
  List.iter
    (fun (key,_,doc) ->
       let cmd = String.sub key 1 (String.length key - 1) in  (* w/o leading "-" *)
       Buffer.add_string b "  ";
       Buffer.add_string b cmd;
       Buffer.add_string b " ";
       Buffer.add_string b doc;
       Buffer.add_string b "\n";
    )
    spec;
  Buffer.contents b
;;


let parse_as_command argv spec =
  let l = Array.length argv in
  let cmd = argv.(1) in
  match cmd with
    | "help" -> raise (Arg.Help (command_help spec))
    | "exit" -> raise End_of_file
    | _ ->
	let (_,cmd_spec,_) =
	  try List.find (fun (n,_,_) -> n = "-" ^ cmd) spec
	  with Not_found -> raise (Arg.Bad ("Unknown command: " ^ cmd)) in
	( match cmd_spec with
	    | Arg.Unit f ->
		if l > 2 then raise (Arg.Bad "Too many arguments");
		f()
	    | Arg.String f ->
		if l < 3 then raise (Arg.Bad "Too few arguments");
		let s = String.concat " " (Array.to_list (Array.sub argv 2 (l - 2))) in
		f s
	    | Arg.Int f ->
		if l < 3 then raise (Arg.Bad "Too few arguments");
		if l > 3 then raise (Arg.Bad "Too many arguments");
		let k =
		  try int_of_string argv.(2)
		  with _ -> raise(Arg.Bad ("Not an integer number: " ^ argv.(2))) in
		f k
	    | _ ->
		assert false
	)
;;


exception Nothing_to_do of string

let interpret ?(toploop=false) argv client_opt id_opt =
  let mode = ref `None in
  let umlid = ref id_opt in

  let current = ref 0 in

  let spec =
    Arg.align
      [ "-info", 
	Arg.Unit (fun () -> mode := `Info), 
	" Show information about the monitor/VM";

	"-tail",
	Arg.String (fun s ->
		      if Str.string_match tail_re s 0 then (
			let file = Str.matched_group 1 s in
			let n =
			  try Int32.of_string(Str.matched_group 3 s)
			  with Not_found -> 20l in
			mode := `Tail(file,n)
		      )
		      else
			raise (Arg.Bad "Invalid tail format")),
	"<file>[,<n>] Show the last <n> lines of <file> (terminal name or 'monitor')";
	
	"-connect", 
	Arg.String (fun s -> mode := `Connect s),
	"<con> Connect with console <con> in an interactive session";
	
	"-force-connect", 
	Arg.String (fun s -> mode := `Force_connect s),
	"<con> Force connection with console <con>, breaking any existing connection";
	
	"-connect-with", 
	Arg.String (fun s ->
		      if Str.string_match connwith_re s 0 then
			let name = Str.matched_group 1 s in
			let filename = Str.matched_group 2 s in
			let rd_flag, wr_flag =
			  match Str.matched_group 3 s with
			    | "r" -> (true, false)
			    | "w" -> (false, true)
			    | "rw" -> (true, true)
			    | _ -> assert false in
			mode := `Connwith(name,filename,rd_flag,wr_flag)
		      else
			raise(Arg.Bad "Bad argument of -connect-with")),
	"<con>=<file>:rw Connect console <con> with <file> in given mode";
	
	"-disconnect",
	Arg.String (fun s -> mode := `Disconnect s),
	"<con> Disconnect console <con>, if connected";

	"-start",
	Arg.Unit (fun () -> mode := `Start),
	" Start the VM";

	"-start-with",
	Arg.String (fun s -> mode := `Start_with s),
	"<arg> Start the VM with additional kernel argument <arg>";

	"-suspend",
	Arg.Unit (fun () -> mode := `Suspend),
	" Suspend the VM";

	"-continue",
	Arg.Unit (fun () -> mode := `Continue),
	" Continue the suspended VM";

	"-halt",
	Arg.Unit (fun () -> mode := `Halt),
	" Halt the VM (emulating a hard crash)";

	"-ctrlaltdel",
	Arg.Unit (fun () -> mode := `Ctrlaltdel),
	" Send CTRL-ALT-DEL to the VM";

	"-kill",
	Arg.Unit (fun () -> mode := `Kill),
	" Kill the VM process";

	"-sysrq",
	Arg.String (fun s -> mode := `Sysrq s),
	"<letter> Send a system request to the VM";

	"-rotate",
	Arg.Unit (fun () -> mode := `Rotate),
	" Reopen the log files";

	"-wait",
	Arg.Int (fun n -> mode := `Wait n),
	"<n> Wait until the VM exits, at most <n> seconds";

	"-port",
	Arg.Unit (fun () -> mode := `Port),
	" Display the port number of the monitor";

	"-set",
	Arg.String (fun s -> 
		      if Str.string_match set_re s 0 then
			let name = Str.matched_group 1 s in
			let value = Str.matched_group 2 s in
			mode := `Set(name,value)
		      else
			raise(Arg.Bad "Bad argument of -set")),
	"<param>=<value> Set the parameter to the value";

	"-del",
	Arg.String (fun s -> mode := `Del s),
	"<param> Delete this parameter (if possible)";

	"-disk-create",
	Arg.String (fun s -> 
		      if Str.string_match disk_create_re s 0 then (
			let name = Str.matched_group 1 s in
			let size = Int32.of_string (Str.matched_group 2 s) in
			let sparse =
			  try Str.matched_group 3 s <> "" with Not_found -> false in
			let options =
			  if sparse then [| "sparse" |] else [| |] in
			mode := `Disk_create(name,size,options)
		      )
		      else
			raise(Arg.Bad "Bad argument of -disk-create")
		   ),
	"<n>:<s>[:sparse] Create disk <n> with <s> megabytes, optionally sparse";

	"-disk-create-cow",
	Arg.String (fun s -> 
		      if Str.string_match disk_create_cow_re s 0 then (
			let cow_name = Str.matched_group 1 s in
			let bak_name = Str.matched_group 2 s in
			mode := `Disk_create_cow(bak_name,cow_name,[| |])
		      )
		      else
			raise(Arg.Bad "Bad argument of -disk-create-cow")
		   ),
	"<cow>:<b> Create copy-on-write disk <cow> backed by <b>";

	"-disk-delete",
	Arg.String (fun s -> mode := `Disk_delete (s,[| |])),
	"<name> Delete the disk <name>";

	"-disk-copy",
	Arg.String (fun s -> 
		      if Str.string_match disk_copy_re s 0 then (
			let src_name = Str.matched_group 1 s in
			let dst_name = Str.matched_group 2 s in
			mode := `Disk_copy(src_name,dst_name,[| |])
		      )
		      else
			raise(Arg.Bad "Bad argument of -disk-copy")
		   ),
	"<src>:<dst> Copy disk <src> to new disk <dst>";

	"-disk-resize",
	Arg.String (fun s -> 
		      if Str.string_match disk_resize_re s 0 then (
			let name = Str.matched_group 1 s in
			let size = Int32.of_string (Str.matched_group 2 s) in
			let sparse =
			  try Str.matched_group 3 s <> "" with Not_found -> false in
			let options = 
			  if sparse then [| "sparse" |] else [||] in
			mode := `Disk_resize(name,size,options)
		      )
		      else
			raise(Arg.Bad "Bad argument of -disk-resize")
		   ),
	"<n>:<s>[:sparse] Resize disk <n> to <s> megabytes, optionally sparse";

	"-change-password",
	Arg.Unit (fun () -> mode := `Change_password),
	" Change user password";

	"-site",
	Arg.String (fun s -> mode := `Site s),
	"<args> Execute the site command with <args>";

	"-version", Arg.Unit (fun s -> mode := `Version), 
	" Print version number";
      ] in

  if toploop then
    parse_as_command argv spec
  else
    Arg.parse_argv
      ~current 
      argv
      spec
      (fun s ->
	 match !umlid with
	   | None -> umlid := Some s
	   | Some _ -> raise(Arg.Bad ("Don't know what to do with: " ^ s)))
      "usage: umlwatch [options] <umlid>";
      
  let id =
    match !umlid with
      | None ->
	  if !mode <> `Version then
	    failwith "Name of the VM to contact is missing!"
	  else
	    ""
      | Some id ->
	  id
  in

  let client = lazy(match client_opt with
		      | Some c -> c
		      | None -> contact id) in
  match !mode with
    | `None        -> raise (Nothing_to_do id)
    | `Info        -> info (Lazy.force client)
    | `Connect con -> connect (Lazy.force client) con
    | `Force_connect con -> force_connect (Lazy.force client) con
    | `Connwith args -> do_connwith args (Lazy.force client)
    | `Disconnect con ->
 do_proc Cl.disconnect ({ ct_name = con; ct_id = 0l }, true) (Lazy.force client);
    | `Start       -> do_proc Cl.start [| |] (Lazy.force client)
    | `Start_with s -> let args = Array.of_list(Str.split (Str.regexp "[ \t\r\n]") s) in
                      do_proc Cl.start args (Lazy.force client)
    | `Suspend     -> do_proc Cl.stop () (Lazy.force client)
    | `Continue    -> do_proc Cl.continue () (Lazy.force client)
    | `Halt        -> do_proc Cl.halt () (Lazy.force client)
    | `Ctrlaltdel  -> do_proc Cl.ctrlaltdel () (Lazy.force client)
    | `Kill        -> do_proc Cl.kill () (Lazy.force client)
    | `Sysrq l     -> do_proc Cl.sysrq l (Lazy.force client)
    | `Rotate      -> do_proc Cl.rotate () (Lazy.force client)
    | `Wait n      -> vm_wait n (Lazy.force client)
    | `Port        -> show_port (Lazy.force client)
    | `Disk_create arg     ->  do_proc Cl.disk_create arg (Lazy.force client)
    | `Disk_create_cow arg ->  do_proc Cl.disk_create_cow arg (Lazy.force client)
    | `Disk_delete arg     -> do_proc Cl.disk_delete arg (Lazy.force client)
    | `Disk_copy arg       -> do_proc Cl.disk_copy arg (Lazy.force client)
    | `Disk_resize arg     -> do_proc Cl.disk_resize arg (Lazy.force client)
    | `Set arg     -> set_parameter arg (Lazy.force client)
    | `Del arg     -> del_parameter arg (Lazy.force client)
    | `Change_password     -> change_pw id
    | `Tail arg            -> tail arg (Lazy.force client)
    | `Site args           -> site_cmd args (Lazy.force client)
    | `Version             -> version()
;;


exception Error

let protect f arg =
  try
    f arg
  with
    | Nothing_to_do _ ->
	()
    | Arg.Help msg ->
	prerr_endline msg
    | Arg.Bad msg ->
	prerr_endline msg; raise Error
    | End_of_file ->
	raise End_of_file
    | Failure msg ->
	prerr_endline ("umlwatch: " ^ msg); raise Error
    | Sys_error msg ->
	prerr_endline ("umlwatch: " ^ msg); raise Error
    | Unix.Unix_error(e,_,_) ->
	prerr_endline ("umlwatch: " ^ Unix.error_message e); raise Error
    | Cannot_connect_term id ->
	prerr_endline ("umlwatch: Cannot connect to terminal " ^ id); raise Error
    | Rpc.Rpc_server Rpc.System_err ->
	prerr_endline ("umlwatch: Server replied with 'system error' - look into log files for more information"); raise Error
    | Rpc.Rpc_server Rpc.Auth_failed ->
	prerr_endline ("umlwatch: Authentication failed"); raise Error
    | Rpc.Rpc_server (Rpc.Unavailable_version _) ->
	prerr_endline ("umlwatch: The server does not support this version of the umlmon protocol"); raise Error
    | err ->
	prerr_endline ("umlwatch: Caught exception " ^ Printexc.to_string err); raise Error
;;


let toploop id =
  version();
  let client = contact id in
  try
    while true do
      printf "%s> " id;
      flush stdout;
      let line = read_line() in
      let words = Str.split (Str.regexp "[ \t\r\n]+") line in
      ( match words with
	  | [] -> ()
	  | cmd :: args ->
	      let line_argv = Array.of_list ("umlwatch" :: cmd :: args) in
	      ( try
		  protect
		    (interpret ~toploop:true line_argv (Some client)) (Some id)
		with
		    Error -> ()
	      )
      )
    done
  with
    | End_of_file -> ()
;;


let main () =
  try
    interpret Sys.argv None None
  with
      Nothing_to_do id ->
	toploop id
;;


try
  Sys.set_signal Sys.sigpipe Sys.Signal_ignore;
  protect main();
with
    Error -> exit 2
;;

